# -*- coding: utf-8 -*-

"""
file: pymorphy.py

Description: this module allows to use pymorphy2 for lemmatizing and POS tagging russian.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse

from sem.processors import PymorphyProcessor


def main(argv=None):
    pymorphy(parser.parse_args(argv))


def pymorphy(
    infile,
    outfile,
    token_field,
    ienc=None,
    oenc=None,
    enc="utf-8",
    log_level="WARNING",
    log_file=None,
):
    token_field = int(token_field or 0)
    processor = PymorphyProcessor()

    with open(infile, encoding=ienc or enc) as input_stream:
        with open(outfile, "w", encoding=oenc or enc) as output_stream:
            for line in input_stream:
                parts = line.strip().split()
                if parts:
                    token = parts[token_field]
                    analyzed = processor._morph.parse(token)
                    parts.append(analyzed[0].normal_form)
                    parts.append(str(analyzed[0].tag.POS or analyzed[0].tag).split(',')[0])
                    output_stream.write("\t".join(parts))
                output_stream.write("\n")


parser = argparse.ArgumentParser("Annotate file with pymorphy2.")

parser.add_argument("infile",
                    help="The input file (CoNLL format)")
parser.add_argument("outfile",
                    help="The output file (CoNLL format)")
parser.add_argument("-t", "--token-field",
                    help="The token field. From command line, the token column.")
parser.add_argument("--input-encoding", dest="ienc",
                    help="Encoding of the input (default: UTF-8)")
parser.add_argument("--output-encoding", dest="oenc",
                    help="Encoding of the input (default: UTF-8)")
parser.add_argument("--encoding", dest="enc", default="UTF-8",
                    help="Encoding of both the input and the output (default: UTF-8)")
parser.add_argument("-l", "--log", dest="log_level",
                    choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"), default="WARNING",
                    help="Increase log level (default: %(default)s)")
parser.add_argument("--log-file", dest="log_file",
                    help="The name of the log file")



# -*- coding:utf-8 -*-

"""
file: label_consistency.py

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse

from sem.storage import (chunk_annotation_from_sentence, compile_multiword)
from sem.importers import read_conll
from sem.processors import (overriding_label_consistency, non_overriding_label_consistency)


def main(argv=None):
    label_consistency(**vars(parser.parse_args(argv)))


def label_consistency(
    infile,
    outfile,
    token_column=0,
    tag_column=-1,
    label_consistency="overriding",
    ienc=None,
    oenc=None,
    enc="utf-8",
):
    ienc = ienc or enc
    oenc = oenc or enc

    entities = {}
    counts = {}
    for p in read_conll(infile, ienc):
        G = chunk_annotation_from_sentence(p, column=tag_column)
        for entity in G:
            id = entity.value
            form = " ".join(
                [p.feature(token_column)[index] for index in range(entity.lb, entity.ub)]
            )
            if form not in counts:
                counts[form] = {}
            if id not in counts[form]:
                counts[form][id] = 0
            counts[form][id] += 1

    for form, count in counts.items():
        if len(count) == 1:
            entities[form] = list(count.keys())[0]
        else:
            best = sorted(count.keys(), key=lambda x: -count[x])[0]
            entities[form] = best

    if label_consistency == "non-overriding":
        feature = non_overriding_label_consistency
    else:
        feature = overriding_label_consistency

    trie = compile_multiword(entities)
    with open(outfile, "w", encoding=oenc) as output_stream:
        for p in read_conll(infile, ienc):
            p.add(feature(p, entities, trie.data, token_column, tag_column), tag_column)
            for token in zip(*[p.feature(key) for key in p.keys()]):
                output_stream.write(("\t".join(token)) + "\n")
            output_stream.write("\n")


parser = argparse.ArgumentParser("Broadcasts annotations based on form.")

parser.add_argument("infile", help="The input file (CoNLL format)")
parser.add_argument("outfile", help="The output file")
parser.add_argument(
    "-t", "--token-column", type=int, default=0, help="Token column"
)
parser.add_argument(
    "-c", "--tag-column", type=int, default=-1, help="Tagging column"
)
parser.add_argument(
    "--label-consistency",
    choices=("non-overriding", "overriding"),
    default="overriding",
    help="Non-overriding leaves CRF's annotation as they are,"
    " overriding label_consistency erases them if it finds a longer one"
    " (default=%(default)s).",
)
parser.add_argument("--input-encoding", dest="ienc", help="Encoding of the input (default: utf-8)")
parser.add_argument("--output-encoding", dest="oenc", help="Encoding of the input (default: utf-8)")
parser.add_argument(
    "-e",
    "--encoding",
    dest="enc",
    default="utf-8",
    help="Encoding of both the input and the output (default: utf-8)",
)
parser.add_argument(
    "-v",
    "--verbose",
    dest="verbose",
    action="store_true",
    help="Writes feedback during process (default: no output)",
)



# -*- coding: utf-8 -*-

"""
file: clean.py

Description: remove unwanted fields from a CoNLL-formatted input.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse

from sem.util import ranges_to_set
import sem.logger


def main(argv=None):
    clean(**vars(parser.parse_args(argv)))


def clean(
    infile, outfile, ranges, ienc=None, oenc=None, enc="utf-8", log_level="WARNING", log_file=None
):
    """
    Cleans a CoNLL-formatted file, removing fields at given indices.

    Parameters
    ----------
    infile : str
        the name of the file to clean.
    outfile : str
        the name of the output file, where some columns have been removed.
    ranges : str
        the fields to remove. Fields is a coma-separated list of indices
        or ranges of indices using a python format (ie: "lo:hi").
    """

    if log_file is not None:
        sem.logger.addHandler(sem.logger.file_handler(log_file))
    sem.logger.setLevel(log_level)

    ienc = ienc or enc
    oenc = oenc or enc

    allowed = ranges_to_set(
        ranges,
        len(open(infile, "rU", encoding=ienc).readline().strip().split()),
        include_zero=True,
    )
    max_abs = 0
    for element in allowed:
        element = abs(element) + (1 if element > 0 else 0)
        max_abs = max(max_abs, element)
    nelts = len(open(infile, "rU", encoding=ienc).readline().strip().split())

    if nelts < max_abs:
        sem.logger.error(
            'asked to keep up to {0} field(s), yet only {1} are present in the "{2}"'.format(
                max_abs, nelts, infile
            )
        )
        raise RuntimeError(
            'asked to keep up to {0} field(s), yet only {1} are present in the "{2}"'.format(
                max_abs, nelts, infile
            )
        )

    sem.logger.info('cleaning "{0}"'.format(infile))
    sem.logger.info(
        "keeping columns: {0}".format(", ".join([str(s) for s in sorted(allowed)]))
    )
    sem.logger.info('writing "{0}"'.format(outfile))

    with open(outfile, "w", encoding=oenc) as output_stream:
        for line in open(infile, "rU", encoding=ienc):
            line = line.strip().split()
            if line != []:
                tokens = [line[i] for i in range(len(line)) if i in allowed]
                output_stream.write("\t".join(tokens))
            output_stream.write("\n")

    sem.logger.info("done")


parser = argparse.ArgumentParser("Remove unwanted columns from CoNLL-formatted file.")

parser.add_argument("infile", help="The input file")
parser.add_argument("outfile", help="The output file ")
parser.add_argument("ranges", help="The ranges of indexes to keep in the file.")
parser.add_argument("--input-encoding", dest="ienc", help="Encoding of the input (default: UTF-8)")
parser.add_argument("--output-encoding", dest="oenc", help="Encoding of the input (default: UTF-8)")
parser.add_argument(
    "--encoding",
    dest="enc",
    default="UTF-8",
    help="Encoding of both the input and the output (default: UTF-8)",
)
parser.add_argument(
    "-l",
    "--log",
    dest="log_level",
    choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
    default="WARNING",
    help="Increase log level (default: critical)",
)
parser.add_argument("--log-file", dest="log_file", help="The name of the log file")



# -*- coding: utf-8 -*-

"""
file: pipeline.py

Description: a file for basic stuff related to pipelines. This file was
created to workaround a cycle dependency between 'sem.util' and 'sem.processors'
when trying to move the 'load_workflow' method in the 'sem.util' module.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import pathlib
import dill
import configparser
import warnings

try:
    from xml.etree import cElementTree as ET
except ImportError:
    from xml.etree import ElementTree as ET

import sem.logger
import sem.processors
from sem.util import check_model_available


class Pipeline(sem.processors.Processor):
    """A pipeline. It will basically chain processors. We will consider that a
    pipeline is a kind of processor since it will do basically the same job.
    It will make everything more pluggable.
    """

    def __init__(self, pipes, pipeline_mode="all", **kwargs):
        super(Pipeline, self).__init__(**kwargs)

        self._pipes = pipes
        self.pipeline_mode = pipeline_mode

    def __iter__(self):
        for pipe in self._pipes:
            yield pipe

    def __len__(self):
        return len(self._pipes)

    def __getitem__(self, index):
        return self._pipes[index]

    @property
    def pipes(self):
        return self._pipes

    def append(self, pipe):
        self._pipes.append(pipe)

    def remove(self, pipe):
        self._pipes.remove(pipe)

    def process_document(self, document, **kwargs):
        for pipe in self._pipes:
            if pipe.check_mode(self.pipeline_mode):
                pipe.process_document(document, **kwargs)
            else:
                sem.logger.info("pipe %s not executed", pipe)
        return document  # allows multiprocessing


def load_workflow(workflow, force_format="default", pipeline_mode="all"):
    """Load a SEM workflow from a file.

    Parameters
    ----------
    workflow : str
        the path to the file.
    force_format : str ["default"]
        if "default", use the normal format defined in workflow file. Otherwise,
        use force_format.
    """

    tree = ET.parse(str(pathlib.Path(workflow).resolve()))
    root = tree.getroot()
    xmlpipes, xmloptions = list(root)

    options = configparser.RawConfigParser()
    exporter = None
    couples = {}
    for xmloption in xmloptions:
        section = xmloption.tag
        options.add_section(section)
        attribs = {}
        for key, val in xmloption.attrib.items():
            key = key.replace("-", "_")
            try:
                attribs[key] = sem.util.str2bool(val)
            except ValueError:
                attribs[key] = val
        for key, val in attribs.items():
            options.set(section, key, val)
        if xmloption.tag == "export":
            couples = dict(options.items("export"))
            export_format = couples["format"]
            if force_format is not None and force_format != "default":
                sem.logger.info("using forced format: {0}".format(force_format))
                export_format = force_format
            exporter = sem.exporters.get_exporter(export_format)(**couples)

    pipes = []
    for xmlpipe in xmlpipes:
        if xmlpipe.tag == "export":
            continue

        arguments = {}
        arguments["expected_mode"] = pipeline_mode
        for key, value in xmlpipe.attrib.items():
            path = pathlib.Path(value)
            user_path = path.expanduser()
            if path != user_path:  # path startswith "~"
                value = str(user_path)
            elif str(path).startswith("../") or str(path).startswith("./"):
                value = str((pathlib.Path(workflow).parent / path).resolve())
            arguments[key.replace("-", "_")] = value
        if list(xmlpipe):
            subpipes = list(xmlpipe)
            for pipe in subpipes:
                path = pipe.attrib.get("path")
                if path and path.startswith("../") or path.startswith("./"):
                    pipe.attrib["path"] = str((pathlib.Path(workflow).parent / path).resolve())
                if pipe.attrib.get("priority", "top-down") == "top-down":
                    subpipes = subpipes[::-1]
            arguments["xmllist"] = subpipes
        for section in options.sections():
            if section == "export":
                continue
            for key, value in options.items(section):
                if key not in arguments:
                    arguments[key] = value
                else:
                    sem.logger.warning("Not adding already existing option: {0}".format(key))
        sem.logger.info("loading {0}".format(xmlpipe.tag))
        pipes.append(sem.processors.build_processor(xmlpipe.tag, **arguments))
    pipeline = Pipeline(pipes, pipeline_mode=pipeline_mode)

    return pipeline, options, exporter, couples


def load_pipeline(path, auto_download=True):
    warnings.filterwarnings("always", category=DeprecationWarning)
    warnings.warn("'load_pipeline' is deprecated, use 'load' instead", DeprecationWarning)
    warnings.filterwarnings("default", category=DeprecationWarning)
    return load(path, auto_download=auto_download)


def load(path, auto_download=True):
    """Load a serialized (potentially .tar.gz) pipeline.
    Pipelines are serialized using dill.
    If the pipeline does not exist and auto_download is set to True,
    SEM will attempt to download the pipeline and load it.

    Parameters
    ----------
    path : str or pathlib.Path
        The path (relative or absolute) to the pipeline.
        SEM will look for the file in the following folders:
        "." and sem.PIPELINE_DIR (defined in sem.__init__.py)
    auto_download : Boolean (default: True)
        Download the pipeline if not found on the computer.

    Returns
    -------
    sem.pipeline.Pipeline
        A SEM pipeline ready to process some documents.
    """
    def download_pipeline(path):
        import os
        import urllib.request

        outfile_path = sem.SEM_PIPELINE_DIR / path.with_suffix(".tar.gz")
        url = sem.SEM_RESOURCE_BASE_URL.format(
            branch="main", kind="pipelines", name=str(path), extension=".tar.gz"
        )
        sem.logger.info("downloading %s...", url)
        response = urllib.request.urlopen(url)
        data = response.read()
        try:
            os.makedirs(outfile_path.parent)
        except FileExistsError:
            pass
        with open(outfile_path, "wb") as output_stream:
            output_stream.write(data)

    path = pathlib.Path(path)
    pipeline = None
    for folder in [pathlib.Path(), sem.SEM_PIPELINE_DIR]:
        actual_path = folder / path
        try:
            check_model_available(actual_path)
            sem.logger.info("Loading %s", str(actual_path))
            with open(actual_path, "rb") as input_stream:
                pipeline = dill.load(input_stream)
        except FileNotFoundError:
            pass

    if pipeline is None:
        if auto_download:
            download_pipeline(path)
            actual_path = sem.SEM_PIPELINE_DIR / path
            check_model_available(actual_path)  # extract if need be
            sem.logger.info("Loading %s", str(actual_path))
            with open(actual_path, "rb") as input_stream:
                pipeline = dill.load(input_stream)
        else:
            raise FileNotFoundError("Could not find pipeline {}".format(path))

    return pipeline


def save(pipeline, path, mode="w"):
    with open(path, f"{mode}b") as output_stream:
        dill.dump(pipeline, output_stream)


def save_pipeline(pipeline, path):
    warnings.filterwarnings("always", category=DeprecationWarning)
    warnings.warn("'save_pipeline' is deprecated, use 'save' instead", DeprecationWarning)
    warnings.filterwarnings("default", category=DeprecationWarning)
    save(pipeline, path)



# -*- coding: utf-8 -*-

"""
file: importers.py

Description: import methods to handle file format to SEM format

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import glob
import pathlib
import urllib.parse
import urllib.request
import re
import sys

try:
    from xml.etree import cElementTree as ET
except ImportError:
    from xml.etree import ElementTree as ET

from html.parser import HTMLParser

import sem.logger
import sem.util
from sem.storage import Document, SEMCorpus, Corpus, Sentence
from sem.storage import Tag, AnnotationSet
from sem.storage import Segmentation
from sem.storage import Span
from sem.storage import tag_annotation_from_corpus, chunk_annotation_from_corpus


def load(
    filename,
    encoding="utf-8",
    fields=None,
    word_field=None,
    wikinews_format=False,
    strip_html=False,
    tagset_name=None,
    *args,
    **kwargs,
):
    """
    The swiss-knife loader. Guess the input format and load the file using the
    right importer.
    """

    if type(filename) in (Document, SEMCorpus):
        sem.logger.info("detected format: SEM XML")
        return filename

    filename = str(filename)  # may be Path
    if filename.startswith("http"):
        sem.logger.info("detected format: HTML")
        return from_url(filename, strip_html=strip_html, wikinews_format=wikinews_format)

    if filename.endswith(".xml"):
        xml = ET.parse(filename)
        root_tag = xml.getroot().tag
        if root_tag == "sem":
            sem.logger.info("detected format: SEM XML")
            return sem_corpus_from_xml(xml)
        elif root_tag == "document":
            sem.logger.info("detected format: SEM XML")
            return sem_document_from_xml(xml)
        elif root_tag == "GateDocument":
            sem.logger.info("detected format: GATE XML")
            return gate_data(xml, pathlib.Path(filename).name)

    path = pathlib.Path(filename)
    ext = path.suffix
    no_ext = str(path.parent / path.stem)
    if (ext == ".ann") or (ext == ".txt" and pathlib.Path(no_ext + ".ann").exists()):
        sem.logger.info("detected format: BRAT")
        return brat_file(
            filename,
            encoding=encoding,
            tagset_name=tagset_name,
            discontinuous=kwargs.get("discontinuous", "split")
        )

    if fields is not None and word_field is not None:
        sem.logger.info("No specific format found, trying CoNLL")
        return conll_file(
            filename,
            fields,
            word_field,
            encoding=encoding,
            taggings=kwargs.get('taggings'),
            chunkings=kwargs.get('chunkings')
        )

    sem.logger.info("No specific format found, defaulting to text format")
    # if everything else fails, just load as text document
    return text_file(filename, encoding=encoding)


def sem_document_from_xml(xml, chunks_to_load=None, load_subtypes=True, type_separator="."):
    if isinstance(xml, str):
        data = ET.parse(xml)
    elif isinstance(xml, ET.ElementTree):
        data = xml
    elif isinstance(xml, type(ET.Element("a"))):  # did not ind a better way to do this
        data = xml
    else:
        raise TypeError("Invalid type for loading XML-SEM document: {0}".format(type(xml)))

    if isinstance(data, ET.ElementTree):
        root = data.getroot()
    elif isinstance(data, type(ET.Element("a"))):
        root = data

    if root.tag == "sem":
        root = list(root)[0]
    elif root.tag != "document":
        raise TypeError("Invalid XML document type for XML-SEM document: {0}".format(root.tag))

    htmlparser = HTMLParser()
    document = Document(root.attrib.get("name", "_DOCUMENT_"))
    for element in list(root):
        if element.tag == "metadata":
            document._metadatas = element.attrib
        elif element.tag == "content":
            document.content = htmlparser.unescape(element.text)
        elif element.tag == "segmentations":
            for segmentation in list(element):
                spans = [
                    Span(
                        lb=int(span.attrib.get("start", span.attrib["s"])),
                        ub=0,
                        length=int(span.attrib.get("length", span.attrib["l"])),
                    )
                    for span in list(segmentation)
                ]
                reference = segmentation.get("reference", None)
                if reference:
                    reference = document.segmentation(reference)
                document.add_segmentation(
                    Segmentation(segmentation.attrib["name"], spans=spans, reference=reference)
                )
        elif element.tag == "annotations":
            for annotation in list(element):
                tags = []
                for tag in list(annotation):
                    value = tag.attrib.get("value", tag.attrib["v"])
                    if not load_subtypes:
                        value = value.strip(type_separator).split(type_separator)[0]
                    tags.append(
                        Tag(
                            value=value,
                            lb=int(tag.attrib.get("start", tag.attrib["s"])),
                            ub=0,
                            length=int(tag.attrib.get("length", tag.attrib["l"])),
                        )
                    )
                reference = annotation.get("reference", None)
                if reference:
                    reference = document.segmentation(reference)
                annotation = AnnotationSet(annotation.attrib["name"], reference=reference)
                annotation.annotations = tags
                document.add_annotationset(annotation)

    if document.segmentation("tokens") and document.segmentation("sentences"):
        document.corpus.from_segmentation(
            document.content,
            document.segmentation("tokens"),
            document.segmentation("sentences"),
        )

        if chunks_to_load is not None:
            for chunk_to_load in chunks_to_load:
                cur_annot = document.annotationset(chunk_to_load)
                if cur_annot and cur_annot.reference is None:
                    document.set_reference(cur_annot.name, "tokens")
                i = 0
                sent_iter = iter(document.corpus)
                shift = 0
                present = set([(a.lb, a.ub) for a in cur_annot])
                for sentence in document.segmentation("sentences"):
                    sent = next(sent_iter)
                    annots = []
                    while i < len(cur_annot) and cur_annot[i].ub <= sentence.ub:
                        annots.append(cur_annot[i])
                        if tuple([cur_annot[i].lb, cur_annot[i].ub]) not in present:
                            raise Exception
                        i += 1
                    l1 = ["O" for _ in range(len(sentence))]
                    for annot in annots:
                        l1[annot.lb - shift] = "B-{0}".format(annot.value)
                        for j in range(annot.lb + 1 - shift, annot.ub - shift):
                            l1[j] = "I-{}".format(annot.value)
                    for j in range(len(l1)):
                        sent[j]["NER"] = l1[j]
                    shift += len(sentence)
                document.corpus.fields.append(chunk_to_load)

    return document


def sem_corpus_from_xml(xml, chunks_to_load=None, load_subtypes=True, type_separator="."):
    if isinstance(xml, str):
        data = ET.parse(xml)
    elif isinstance(xml, ET.ElementTree):
        data = xml
    elif isinstance(xml, type(ET.Element("a"))):  # did not ind a better way to do this
        data = xml
    else:
        raise TypeError("Invalid type for loading XML-SEM document: {0}".format(type(xml)))

    root = data.getroot()
    if root.tag != "sem":
        raise ValueError("Not sem xml file type: '{0}'".format(root.tag))
    doc_list = []
    for document in list(root):
        doc_list.append(Document.from_xml(document))
    return SEMCorpus(doc_list)


def add_document(self, document):
    ok = not any([d.name == document.name for d in self.documents])
    if ok:
        self._documents.append(document)


def text_file(filename, encoding="utf-8"):
    """
    Read a text file.
    """
    with open(filename, "r", encoding=encoding, newline="") as input_stream:
        content = input_stream.read().replace("\r", "")

    return Document(
        pathlib.Path(filename).name,
        content=content,
        encoding=encoding,
    )


def read_conll(name, encoding, fields=None, cleaner=str.strip, splitter=str.split):
    """
    Iterate through a CoNLL-formatted file and provide sentences represented as
    either a list of list (if fields is None) or a list of dict (otherwise).
    """
    clean = cleaner or (lambda x: x)
    split = splitter or (lambda x: x)
    if fields:
        def to_data(x, y): return {key: val for key, val in zip(y, x)}
    else:
        def to_data(x, y): return {index: item for index, item in enumerate(x)}

    with open(name, "r", encoding=encoding, newline="") as input_stream:
        paragraph = []
        for line in input_stream:
            line = clean(line)
            if line != "":
                paragraph.append(split(line))
            elif paragraph != []:
                paragraph = [list(item) for item in zip(*paragraph)]
                yield Sentence(to_data(paragraph, fields))
                del paragraph[:]
        if paragraph != []:
            paragraph = [list(item) for item in zip(*paragraph)]
            yield Sentence(to_data(paragraph, fields))


def conll_file(filename, fields, word_field, encoding="utf-8", taggings=None, chunkings=None):
    """
    Read CoNLL-formatted text from a file.
    """
    name = pathlib.Path(filename).name
    sents = [sent for sent in read_conll(filename, encoding, fields)]
    fields = fields or sorted(sents[0].keys())  # no fields = we infer a "list" (dict with 0-index)
    if word_field is None:
        word_field = fields[0]
    return conll_data(
        name,
        Corpus(fields=fields, sentences=sents),
        word_field,
        encoding="utf-8",
        taggings=taggings,
        chunkings=chunkings
    )


def conll_data(name, corpus, word_field, encoding="utf-8", taggings=None, chunkings=None):
    """
    Create a Document from CoNLL-formatted data (SEM Corpus).
    """
    character_index = 0
    sentence_index = 0
    contents = []
    word_spans = []
    sentence_spans = []
    for sentence in corpus.sentences:
        contents.append(sentence.feature(word_field))
        for word in contents[-1]:
            word_spans.append(Span(character_index, character_index + len(word)))
            character_index += len(word) + 1
        sentence_spans.append(Span(sentence_index, sentence_index + len(sentence)))
        sentence_index += len(sentence)
    document = Document(
        name,
        content="\n".join([" ".join(content) for content in contents]),
        encoding=encoding,
        corpus=corpus
    )
    document.add_segmentation(Segmentation("tokens", spans=word_spans))
    document.add_segmentation(
        Segmentation(
            "sentences", reference=document.segmentation("tokens"), spans=sentence_spans[:]
        )
    )
    for tagging in taggings or []:
        document.add_annotationset(
            tag_annotation_from_corpus(
                document._corpus,
                tagging,
                tagging,
                reference=document.segmentation("tokens"),
                strict=True,
            )
        )
    for chunking in chunkings or []:
        document.add_annotationset(
            chunk_annotation_from_corpus(
                document._corpus,
                chunking,
                chunking,
                reference=document.segmentation("tokens"),
                strict=True,
            )
        )
    return document


def from_url(url, strip_html=False, wikinews_format=False, keep_offsets=True, **kwargs):
    """
    Load a SEM Document using an URL.

    Parameters
    ----------
    url : str
        The URL to use.
    strip_html : boolean [False]
        Do we strip the HTML from the document's content?
    wikinews_format : boolean [False]
        Do we use the wikinews format?
    **kwargs : kwargs
        other arguments
    """
    url = url.strip()

    if not url:
        return None

    strip_html |= wikinews_format  # wikinews format is always stripped
    keep_offsets &= not wikinews_format  # wikinews format never keeps offsets

    charset = re.compile(b'charset="(.+?)"')
    escaped_url = "".join([(urllib.parse.quote(c) if ord(c) > 127 else c) for c in url])
    escaped_url = escaped_url.replace("%2525", "%25")
    escaped_url = escaped_url.replace('"', "&quot;")

    content = ""
    f = urllib.request.urlopen(escaped_url)
    content = f.read()
    f.close()
    match = charset.search(content)
    if match is not None:
        encoding = match.group(1).decode("utf-8") or "utf-8"
    else:
        encoding = "utf-8"
    content = content.decode(encoding)

    regex = re.compile("^.+?[^/]/(?=[^/])", re.M)
    parts = regex.findall(escaped_url)
    base_url = (escaped_url[:] + "/" if parts == [] else parts[0])

    content = content.replace('="//', '="http://')
    content = content.replace('="/', '="{0}'.format(base_url))
    content = content.replace('=\\"//', '=\\"http://')
    content = content.replace('=\\"/', '=\\"{0}'.format(base_url))
    content = content.replace("\r", "")
    content = content.replace("</p>", "</p>\n\n")

    if strip_html:
        new_content = sem.util.strip_html(content, keep_offsets=keep_offsets)
    else:
        new_content = content

    if wikinews_format:
        cleaned_content = new_content[: content.index("<h2>")].strip()
    else:
        cleaned_content = new_content

    """if strip_html:
        h = HTMLParser()
        empty_line = re.compile("\n[ \t]+")
        spaces = re.compile("[ \t]+")
        newlines = re.compile("\n{2,}")
        cleaned_content = h.unescape(cleaned_content)
        cleaned_content = empty_line.sub("\n", cleaned_content)
        cleaned_content = spaces.sub(" ", cleaned_content)
        cleaned_content = newlines.sub("\n\n", cleaned_content)

    spaces_begin = re.compile("^[ \t]+", re.M)
    spaces_end = re.compile("[ \t]+$", re.M)
    cleaned_content = spaces_begin.sub("", cleaned_content)
    cleaned_content = spaces_end.sub("", cleaned_content)"""

    mime_type = "text/plain" if strip_html else "text/html"
    return Document(
        name=url, content=cleaned_content, original_content=content, mime_type=mime_type
    )


def brat_file(filename, encoding="utf-8", tagset_name=None, discontinuous="split"):
    """
    Load a BRAT file couple. A BRAT file couple is a ".txt" and a ".ann" file.
    If no name is provided for the annotation set, NER is used instead.

    Parameters
    ----------
    filename : str
        the path to the file.
    encoding : str ["utf-8"]
        the encoding of the file.
    tagset_name : str [None]
        the name of the annotation set to load.
    """
    tagset_name = tagset_name or "NER"
    path = pathlib.Path(filename)
    no_ext = str(path.parent / path.stem)
    txt_file = no_ext + ".txt"
    ann_file = no_ext + ".ann"
    if not (pathlib.Path(txt_file).exists() and pathlib.Path(ann_file).exists()):
        raise ValueError("missing either .ann or .txt file")

    document = Document(pathlib.Path(txt_file).name, encoding=encoding, mime_type="text/plain")
    document.content = open(txt_file, "r", encoding=encoding).read().replace("\r", "\n")
    annotations = AnnotationSet(tagset_name)
    for line in open(ann_file, "r", encoding=encoding, newline=""):
        line = line.strip()
        if line != "" and line.startswith("T"):
            parts = line.split("\t")
            value, bounds = parts[1].split(" ", 1)
            if discontinuous == "split":
                for bound in bounds.split(";"):
                    lb, ub = bound.split()
                    lb = int(lb)
                    ub = int(ub)
                    annotations.append(Tag(value=value, lb=lb, ub=ub))
            elif discontinuous == "merge":
                parts = bounds.split()
                lb = int(parts[0])
                ub = int(parts[-1])
                annotations.append(Tag(value=value, lb=lb, ub=ub))
            else:
                raise ValueError(
                    f'Invalid discontinuous annotation handling: "{discontinuous}" (split, merge)'
                )
    annotations.sort()
    document.add_annotationset(annotations)

    return document


def gate_file(filename):
    """
    Load a GATE-formatted file.

    Parameters
    ----------
    filename : str
        the path to the file to load.
    """
    data = ET.parse(filename)
    return gate_data(data, name=pathlib.Path(filename).name)


def gate_data(data, name=None):
    """
    Load a GATE-formatted ElementTree. GATE-formatted data contains the content
    and an annotation set. If no name is provided for the annotation set, NER
    is used instead.

    Parameters
    ----------
    data : str
        the GATE ELementTree to load as a SEM Document.
    name : str [None]
        the name of the annotation set to load.
    """
    document = Document(name or "__DOCUMENT__", mime_type="text/plain")

    textwithnodes = data.findall("TextWithNodes")[0]
    annotation_sets = data.findall("AnnotationSet")

    text_parts = [textwithnodes.text or ""]
    nodes = {}
    for node in list(textwithnodes):
        nodes[int(node.attrib["id"])] = sum([len(part) for part in text_parts])
        text_parts.append(node.tail or "")
    document.content = "".join(text_parts)

    for annotation_set in annotation_sets:
        annotation_name = annotation_set.attrib["Name"]
        sem_annotation = AnnotationSet(annotation_name)
        for annotation in annotation_set:
            lb = nodes[int(annotation.attrib["StartNode"])]
            ub = nodes[int(annotation.attrib["EndNode"])]
            sem_annotation.append(Tag(annotation.attrib["Type"], lb, ub))
        document.add_annotationset(sem_annotation)

    return document


def json_data(data):
    """
    Load Document from json data. The json is just a dict containing the right
    informations.

    Parameters
    ----------
    data : dict
        the json input data
    """
    document = Document(data.get("name", "_DOCUMENT_"), content=data.get("content", ""))
    for key, value in data.get("metadatas", {}).items():
        document.add_metadata(key, value)

    for segmentation_name in data.get("segmentations", {}):
        d = data["segmentations"][segmentation_name]
        spans = [Span(lb=span["s"], ub=0, length=span["l"]) for span in d["spans"]]
        segmentation = Segmentation(
            segmentation_name, spans=spans, reference=d.get("reference", None)
        )
        document.add_segmentation(segmentation)
    for segmentation in document.segmentations:
        if segmentation.reference is not None:
            segmentation.reference = document.segmentation(segmentation.reference)

    for annotation_name in data.get("annotations", {}):
        d = data["annotations"][annotation_name]
        annotations = [
            Tag(value=annotation["v"], lb=annotation["s"], ub=0, length=annotation["l"])
            for annotation in d["annotations"]
        ]
        annotation = AnnotationSet(
            annotation_name,
            reference=document.segmentation(d["reference"]),
            annotations=annotations,
        )
        document.add_annotationset(annotation)


def documents_from_list(name_list, file_format, **opts):
    """
    Create a Document list from a list which may contain either Document objects
    or string objects that need to be globbed.

    Parameters
    ----------
    name_list : list
        the list of "documents". It can be Document object or str with wildcards.
    file_format : str
        The expected file format for documents. Can be "plain", "conll",
        "guess", etc.
    **opts : dict
        options for reading documents.
    """
    documents = []
    names = set()  # document names that were already seen
    for name in name_list:
        if isinstance(name, Document):
            sem.logger.info("Reading %s", name.name)
            if name.name not in names:
                documents.append(name)
                names.add(name.name)
            else:
                sem.logger.info("document %s already found, not adding to the list.", name.name)
        else:
            name = str(name)
            for infile in glob.glob(name) or [name]:
                sem.logger.info("Reading %s", infile)
                if file_format == "text":
                    document = Document(
                        pathlib.Path(infile).name,
                        content=open(infile, "r", encoding="utf-8", newline="")
                        .read()
                        .replace("\r", ""),
                        **opts,
                    )
                elif file_format == "conll":
                    document = conll_file(infile, **opts)
                elif file_format == "html":
                    try:
                        infile = infile.decode(sys.getfilesystemencoding())
                    except Exception:
                        pass
                    document = from_url(infile, **opts)
                elif file_format == "guess":
                    document = load(infile, **opts)
                else:
                    raise ValueError("unknown format: {0}".format(file_format))
                if document.name not in names:
                    documents.append(document)
                    names.add(document.name)
                else:
                    sem.logger.info(
                        "document %s already found, not adding to the list.", document.name
                    )

    return documents



# -*- coding: utf-8 -*-

"""
file: download.py

Description: Download a SEM resource from GitHub.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import os
import pathlib
import tarfile
import urllib.request

import sem
from sem.util import str2bool
import sem.logger


# resources packs are a special case of resource as they represent multiple resources
RESOURCE_PACK = "resource-pack"
RESOURCE_PACKS = "resource-packs"

__kind2name = {
    "dictionary": "dictionaries",
    "workflow": "workflow",
    "model": "models",
    "pipeline": "pipelines",
    "tagset": "tagsets",
    RESOURCE_PACK: RESOURCE_PACKS,
}


def do_overwrite(path, conflict):
    if conflict == "skip":
        return False
    if conflict == "overwrite":
        return True

    answer = input(f"{path} already exists, override? [y/N] ").lower()
    try:
        return str2bool(answer)
    except ValueError:
        return False


def main(argv=None):
    download(**vars(parser.parse_args(argv)))


def download(
    name,
    kind="pipeline",
    branch="main",
    extension=".tar.gz",
    output_dir=sem.SEM_RESOURCE_DIR,
    encoding="utf-8",
    binary_or_text="guess",
    extract=False,
    clean=False,
    conflict="prompt",
):
    """Download a SEM resource from GitHub."""

    log_lvl = sem.logger.level
    sem.logger.setLevel("INFO")

    kind = __kind2name[kind]

    # resource packs are a special case, they contain multiple resources (folders) and can only be
    # a compressed file (.tar.gz) that will be placed and uncompressed in base output folder.
    if kind == RESOURCE_PACKS:
        sem.logger.info("downloading a resource pack: using pre-made configuration...")
        extension = ".tar.gz"
        binary_or_text = "b"
        extract = True
        clean = True
        actual_name = name.replace("/", "_")
        relative_path = pathlib.Path(f"{actual_name}{extension}")
    else:
        relative_path = pathlib.Path(f"{kind}/{name}{extension}")

    outfile_path = pathlib.Path(output_dir) / relative_path

    if outfile_path.exists():
        answer = input(f"{outfile_path} already exists, overwrite? [y/N] ").lower()
        try:
            overwrite = str2bool(answer)
        except ValueError:
            overwrite = False
        if not overwrite:
            return
    else:
        try:
            os.makedirs(outfile_path.parent)
        except FileExistsError:
            pass

    url = sem.SEM_RESOURCE_BASE_URL.format(
        branch=branch, kind=kind, name=name, extension=extension
    )
    sem.logger.info("downloading %s...", url)
    response = urllib.request.urlopen(url)
    data = response.read()
    if binary_or_text == "guess":
        binarytext = ("b" if extension in ["", ".tar.gz"] else "t")
    else:
        binarytext = binary_or_text
    if binarytext == "t":
        data = str(data, encoding)

    with open(outfile_path, "w{}".format(binarytext)) as output_stream:
        output_stream.write(data)

    if extract and extension == ".tar.gz":
        sem.logger.info("extracting %s...", outfile_path)
        with tarfile.open(outfile_path, "r:gz") as tar:
            for name in tar.getnames():
                path = outfile_path.parent / name
                do_skip = path.exists() and not path.is_dir() and not do_overwrite(path, conflict)
                if not do_skip:
                    tar.extract(name, outfile_path.parent)
        if clean:
            sem.logger.info("deleting %s...", outfile_path)
            outfile_path.unlink()

    sem.logger.setLevel(log_lvl)


parser = argparse.ArgumentParser(
    "Download a SEM resource from GitHub: https://github.com/YoannDupont/SEM-resources\n"
    "A resource's name will first contain its lang. For example, base french resource pack's name"
    " is 'fr/base'. Example commands for french resources that illustrate some options:\n"
    "\n"
    "sem download fr/base --kind resource-pack\n"
    "sem download fr/FTB-POS_NER --extract --clean --conflict overwrite\n"
)

parser.add_argument("name", help="The name of the resource (example: fr/FTB-POS_NER)")
parser.add_argument(
    "-k",
    "--kind",
    choices=sorted(__kind2name.keys()),
    default="pipeline",
    help="The kind of resource to download (default: %(default)s)"
)
parser.add_argument(
    "-b",
    "--branch",
    default="main",
    help="The branch from which to download (default: %(default)s)"
)
parser.add_argument(
    "-t",
    "--extension",
    default=".tar.gz",
    help="The extension of the file you want to download (default: %(default)s)"
)
parser.add_argument(
    "-o",
    "--output_dir",
    default=sem.SEM_RESOURCE_DIR,
    help="The base folder where to download the resource (default: %(default)s)"
)
parser.add_argument(
    "-e", "--encoding",
    default="UTF-8",
    help="Encoding of the output file, if non binary (default: %(default)s)",
)
parser.add_argument(
    "--binary-or-text",
    choices=("b", "t", "guess"),
    default="guess",
    help="Is the file a binary, a text or should the program guess? (default: %(default)s)"
)
parser.add_argument(
    "-x",
    "--extract",
    action="store_true",
    help="If the resource is an archive, extract it"
)
parser.add_argument(
    "--clean",
    action="store_true",
    help="If the resource is an archive, delete the archive after extraction"
)
parser.add_argument(
    "--conflict",
    choices=("prompt", "overwrite", "skip"),
    default="prompt",
    help="The action to perform to solve conflicts when extracting file (default: %(default)s)"
)



# -*- coding: utf-8 -*-

"""
file: chunking_evaluate.py

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import os
import collections

from sem.importers import read_conll
from sem.storage import AnnotationSet
from sem.storage import annotation_from_sentence

import sem.importers

CORRECT = "correct output"
TYPE_ERROR = "type error"
BOUNDARY_ERROR = "boundary error"
TYPE_AND_BOUNDARY_ERROR = "type+boundary error"
NOISE_ERROR = "noise error"
SILENCE_ERROR = "silence error"
TP = "true positive"
FP = "false positive"
FN = "false negative"

ERRORS_KINDS = [
    FP, FN, TYPE_ERROR, BOUNDARY_ERROR, TYPE_AND_BOUNDARY_ERROR, NOISE_ERROR, SILENCE_ERROR
]
OUTPUT_KINDS = [CORRECT] + ERRORS_KINDS


def mean(numbers):
    return sum(numbers) / len(numbers)


def precision(d):
    numerator = len(d.get(TP, []))
    denominator = len(d.get(TP, []) + d.get(FP, []))
    if denominator == 0:
        return 0.0
    else:
        return numerator / denominator


def recall(d):
    numerator = len(d.get(TP, []))
    denominator = len(d.get(TP, []) + d.get(FN, []))
    if denominator == 0:
        return 0.0
    else:
        return numerator / denominator


def fscore(P, R, beta=1.0):
    return (1 + (beta ** 2)) * P * R / (((beta ** 2) * P) + R) if P + R != 0.0 else 0.0


def undergeneration(d):
    numerator = len(d.get(SILENCE_ERROR, []))
    denominator = len(d.get(TP, []) + d.get(FN, []))
    if denominator == 0:
        return 0.0
    else:
        return numerator / denominator


def overgeneration(d):
    numerator = len(d.get(NOISE_ERROR, []))
    denominator = len(d.get(TP, []) + d.get(FP, []))
    if denominator == 0:
        return 0.0
    else:
        return numerator / denominator


def substitution(d):
    numerator = len(d.get(FP, [])) - len(d.get(NOISE_ERROR, []))
    denominator = (
        len(d.get(TP, []) + d.get(FP, [])) - len(d.get(NOISE_ERROR, []))
    )
    if denominator == 0:
        return 0.0
    else:
        return numerator / denominator


def get_diff(content, gold, guess, error_kind, context_size=20):
    if error_kind == TYPE_ERROR:
        diff = content[gold.lb - context_size: gold.lb]
        diff += "{{+<{0}>+}} ".format(guess.value)
        diff += "[-<{0}>-] ".format(gold.value)
        diff += content[gold.lb: gold.ub]
        diff += " [-</{0}>-]".format(gold.value)
        diff += " {{+</{0}>+}}".format(guess.value)
        diff += content[gold.ub: gold.ub + context_size]
    elif error_kind == BOUNDARY_ERROR:
        if gold.lb == guess.lb:
            diff = content[gold.lb - context_size: gold.lb]
            diff += "<{0}> ".format(gold.value)
            gold_min = gold.ub < guess.ub
            diff += content[gold.lb: min(gold.ub, guess.ub)]
            diff += (
                " [-</{0}>-]".format(gold.value)
                if min(gold.ub, guess.ub)
                else " {{+</{0}>+}}".format(guess.value)
            )
            diff += content[min(gold.ub, guess.ub): max(gold.ub, guess.ub)]
            diff += (
                " {{+</{0}>+}}".format(guess.value)
                if min(gold.ub, guess.ub)
                else " [-</{0}>-]".format(gold.value)
            )
            diff += content[max(gold.ub, guess.ub): max(gold.ub, guess.ub) + context_size]
        else:
            gold_min = gold.lb < guess.lb
            diff = content[min(gold.lb, guess.lb) - context_size: min(gold.lb, guess.lb)]
            diff += (
                "[-<{0}>-] ".format(gold.value) if gold_min else "{{+<{0}>+}} ".format(guess.value)
            )
            diff += content[min(gold.lb, guess.lb): max(gold.lb, guess.lb)]
            diff += (
                "{{+<{0}>+}} ".format(guess.value) if gold_min else "[-<{0}>-] ".format(gold.value)
            )
            diff += content[max(gold.lb, guess.lb): gold.ub]
            diff += " </{0}>".format(gold.value)
            diff += content[gold.ub: gold.ub + context_size]
    elif error_kind == TYPE_AND_BOUNDARY_ERROR:
        min_lb = (
            gold
            if gold.lb < guess.lb
            else (gold if gold.lb == guess.lb and gold.ub > guess.ub else guess)
        )
        max_lb = gold if min_lb == guess else guess
        min_ub = (
            gold
            if gold.ub < guess.ub
            else (gold if gold.ub == guess.ub and gold.lb > guess.lb else guess)
        )
        max_ub = gold if min_ub == guess else guess
        diff = content[min_lb.lb - context_size: min_lb.lb]
        if min_lb == gold:
            diff += "[-<{0}>-] ".format(gold.value)
            diff += content[min_lb.lb: max_lb.lb]
            diff += "{{+<{0}>+}} ".format(guess.value)
        else:
            diff += "{{+<{0}>+}} ".format(guess.value)
            diff += content[min_lb.lb: max_lb.lb]
            diff += "[-<{0}>-] ".format(gold.value)
        diff += content[max_lb.lb: min_ub.ub]
        if min_ub == gold:
            diff += " [-</{0}>-]".format(gold.value)
            diff += content[min_ub.ub: max_ub.ub]
            diff += " {{+</{0}>+}}".format(guess.value)
        else:
            diff += " {{+</{0}>+}}".format(guess.value)
            diff += content[min_ub.ub: max_ub.ub]
            diff += " [-</{0}>-]".format(gold.value)
        diff += content[max_ub.ub: max_ub.ub + context_size]
    elif error_kind == NOISE_ERROR:
        diff = content[guess.lb - context_size: guess.lb]
        diff += "{{+<{0}>+}} ".format(guess.value)
        diff += content[guess.lb: guess.ub]
        diff += " {{+</{0}>+}}".format(guess.value)
        diff += content[guess.ub: guess.ub + context_size]
    elif error_kind == SILENCE_ERROR:
        diff = content[gold.lb - context_size: gold.lb]
        diff += "[-<{0}>-] ".format(gold.value)
        diff += content[gold.lb: gold.ub]
        diff += " [-</{0}>-]".format(gold.value)
        diff += content[gold.ub: gold.ub + context_size]
    else:
        raise ValueError("Unknown error kind: {0}".format(error_kind))
    return diff.replace("\r", "").replace("\n", " ").replace('"', '\\"')


def main(argv=None):
    evaluate(**vars(parser.parse_args(argv)))


def evaluate(
    infile=None,
    reference_column=None,
    tagging_column=None,
    ienc=None,
    input_format=None,
    reference_file=None,
    annotation_name=None,
    dump=os.devnull,
    context_size=30,
):
    counts = {}
    if input_format == "conll":
        if reference_file:
            raise ValueError("reference_file not handled for CoNLL files")
        L = []
        R = []
        keys = None
        nth = -1
        n_line = 0
        for p in read_conll(infile, ienc):
            nth += 1
            keys = keys or list(p.keys())
            L.extend(annotation_from_sentence(p, column=keys[reference_column], shift=n_line - nth))
            R.extend(annotation_from_sentence(p, column=keys[tagging_column], shift=n_line - nth))
            n_line += len(p) + 1
        document = sem.importers.conll_file(infile, keys, keys[0], encoding=ienc)
        L = AnnotationSet(
            "", annotations=L, reference=document.segmentation("tokens")
        ).get_reference_annotations()
        R = AnnotationSet(
            "", annotations=R, reference=document.segmentation("tokens")
        ).get_reference_annotations()
    elif input_format == "brat":
        document = sem.importers.brat_file(reference_file)
        L = document.annotationset("NER").get_reference_annotations()
        R = sem.importers.brat_file(infile).annotationset("NER").get_reference_annotations()
    elif input_format in ("sem", "SEM"):
        document = sem.importers.sem_document_from_xml(reference_file)
        system = sem.importers.sem_document_from_xml(infile)
        common_annotations = set(document.annotationsets.keys()) & set(system.annotationsets.keys())
        if len(common_annotations) == 1 and annotation_name is None:
            annotation_name = list(common_annotations)[0]
        if annotation_name is None:
            raise RuntimeError("Could not find an annotation set to evaluate: please provide one")
        L = document.annotationset(annotation_name).get_reference_annotations()
        R = system.annotationset(annotation_name).get_reference_annotations()
    else:
        raise RuntimeError("format not handled: {0}".format(input_format))

    len_ref = len(L)
    len_tag = len(R)
    d = collections.defaultdict(list)
    # first pass, removing correct
    i = 0
    while i < len(L):
        LR = L[i]
        j = 0
        while j < len(R):
            RR = R[j]
            if LR == RR:
                del L[i]
                del R[j]
                i -= 1
                d[CORRECT].append([LR, RR])
                d[TP].append(LR)
                break
            j += 1
        i += 1

    # second pass, typing errors
    i = 0
    while i < len(L):
        LR = L[i]
        j = 0
        while j < len(R):
            RR = R[j]
            if LR.value != RR.value and LR.lb == RR.lb and LR.ub == RR.ub:
                del L[i]
                del R[j]
                d[TYPE_ERROR].append([LR, RR])
                d[FN].append(LR)
                d[FP].append(RR)
                break
            j += 1
        i += 1

    # third pass, boundary errors
    i = 0
    while i < len(L):
        LR = L[i]
        j = 0
        while j < len(R):
            RR = R[j]
            if LR.value == RR.value and (
                (LR.lb != RR.lb and LR.ub == RR.ub) or (LR.lb == RR.lb and LR.ub != RR.ub)
            ):
                del L[i]
                del R[j]
                i -= 1
                d[BOUNDARY_ERROR].append([LR, RR])
                d[FN].append(LR)
                d[FP].append(RR)
                break
            j += 1
        i += 1

    # fourth pass, both type and boundary errors
    i = 0
    while i < len(L):
        LR = L[i]
        j = 0
        while j < len(R):
            RR = R[j]
            if (
                LR.value != RR.value
                and (LR.lb != RR.lb and LR.ub == RR.ub)
                or (LR.lb == RR.lb and LR.ub != RR.ub)
            ):
                del L[i]
                del R[j]
                i -= 1
                d[TYPE_AND_BOUNDARY_ERROR].append([LR, RR])
                d[FN].append(LR)
                d[FP].append(RR)
                break
            j += 1
        i += 1

    d[SILENCE_ERROR] = L[:]
    d[NOISE_ERROR] = R[:]
    d[FN].extend(L[:])
    d[FP].extend(R[:])

    entities = set()
    for vals in d.values():
        for val in vals:
            try:
                left, right = val
                entities.add(left.value)
                entities.add(right.value)
            except (AttributeError, TypeError):
                entities.add(val.value)

    with open(dump, "w", encoding="utf-8") as output_stream:
        output_stream.write("error kind\treference entity\toutput entity\tdiff\n")
        for error_kind in (
            TYPE_ERROR,
            BOUNDARY_ERROR,
            TYPE_AND_BOUNDARY_ERROR,
            NOISE_ERROR,
            SILENCE_ERROR,
        ):
            for ex in d[error_kind]:
                if error_kind == NOISE_ERROR:
                    gold = None
                    guess = ex
                elif error_kind == SILENCE_ERROR:
                    gold = ex
                    guess = None
                else:
                    gold = ex[0]
                    guess = ex[1]
                gold_str = (
                    (
                        "{0}:{1}".format(gold.value, document.content[gold.lb: gold.ub])
                        if gold
                        else ""
                    )
                    .replace("\r", "")
                    .replace("\n", " ")
                )
                guess_str = (
                    (
                        "{0}:{1}".format(guess.value, document.content[guess.lb: guess.ub])
                        if guess
                        else ""
                    )
                    .replace("\r", "")
                    .replace("\n", " ")
                )
                diff = get_diff(
                    document.content, gold, guess, error_kind, context_size=context_size
                )
                output_stream.write(
                    "{0}\t{1}\t{2}\t{3}\n".format(error_kind, gold_str, guess_str, diff)
                )

    counts = {}
    for entity in entities:
        sub_d = {}
        sub_d[CORRECT] = [m for m in d[CORRECT] if m[0].value == entity]
        sub_d[TYPE_ERROR] = [
            m for m in d[TYPE_ERROR] if m[0].value == entity or m[1].value == entity
        ]
        sub_d[BOUNDARY_ERROR] = [
            m for m in d[BOUNDARY_ERROR] if m[0].value == entity or m[1].value == entity
        ]
        sub_d[TYPE_AND_BOUNDARY_ERROR] = [
            m for m in d[TYPE_AND_BOUNDARY_ERROR] if m[0].value == entity or m[1].value == entity
        ]
        sub_d[NOISE_ERROR] = [m for m in d[NOISE_ERROR] if m.value == entity]
        sub_d[SILENCE_ERROR] = [m for m in d[SILENCE_ERROR] if m.value == entity]
        sub_d[TP] = [m for m in d[TP] if m.value == entity]
        sub_d[FP] = [m for m in d[FP] if m.value == entity]
        sub_d[FN] = [m for m in d[FN] if m.value == entity]
        counts[entity] = sub_d

    # basic counts
    print("entity\tmeasure\tvalue")
    for entity in sorted(entities):
        for kind in OUTPUT_KINDS:
            print("{0}\t{1}\t{2}".format(entity, kind, len(counts[entity][kind])))
    print("global\treference\t{0}".format(len_ref))
    print("global\ttagging\t{0}".format(len_tag))
    for kind in OUTPUT_KINDS:
        print("global\t{0}\t{1}".format(kind, len(d[kind])))

    # P R F
    precisions = []
    recalls = []
    print()
    print("entity\tmeasure\tvalue")
    for entity in sorted(entities):
        precisions.append(precision(counts[entity]))
        recalls.append(recall(counts[entity]))
        print("{0}\tprecision\t{1:.4f}".format(entity, precisions[-1]))
        print("{0}\trecall\t{1:.4f}".format(entity, recalls[-1]))
        print(
            "{0}\tfscore\t{1:.4f}".format(
                entity, fscore(precision(counts[entity]), recall(counts[entity]))
            )
        )
    print("global\tprecision\t{0:.4f}".format(precision(d)))
    print("global\trecall\t{0:.4f}".format(recall(d)))
    print("global\tfscore\t{0:.4f}".format(fscore(precision(d), recall(d))))
    print("global\tmacro-precision\t{0:.4f}".format(mean(precisions)))
    print("global\tmacro-recall\t{0:.4f}".format(mean(recalls)))
    print("global\tmacro-fscore\t{0:.4f}".format(fscore(mean(precisions), mean(recalls))))

    # over/under generation, substitution
    print()
    print("entity\tmeasure\tvalue")
    for entity in sorted(entities):
        print("{0}\tundergeneration\t{1:.4f}".format(entity, undergeneration(counts[entity])))
        print("{0}\tovergeneration\t{1:.4f}".format(entity, overgeneration(counts[entity])))
        print("{0}\tsubstitution\t{1:.4f}".format(entity, substitution(counts[entity])))
    print("global\tundergeneration\t{0:.4f}".format(undergeneration(d)))
    print("global\tovergeneration\t{0:.4f}".format(overgeneration(d)))
    print("global\tsubstitution\t{0:.4f}".format(substitution(d)))


parser = argparse.ArgumentParser("Get F1-score for tagging using the IOB scheme.")

parser.add_argument("infile", help="The input file (CoNLL format)")
parser.add_argument(
    "-r",
    "--reference-column",
    dest="reference_column",
    type=int,
    default=-2,
    help="Column for reference output (default: %(default)s)",
)
parser.add_argument(
    "-t",
    "--tagging-column",
    dest="tagging_column",
    type=int,
    default=-1,
    help="Column for CRF output (default: %(default)s)",
)
parser.add_argument(
    "-f",
    "--format",
    dest="input_format",
    default="conll",
    help="The input format (default: %(default)s)",
)
parser.add_argument(
    "-a",
    "--annotation-name",
    dest="annotation_name",
    help="The annotation name, useful for some formats, like SEM.",
)
parser.add_argument("-c", "--reference-file", dest="reference_file", help="The comparing file")
parser.add_argument("--input-encoding", dest="ienc", help="Encoding of the input (default: utf-8)")
parser.add_argument(
    "-d",
    "--dump",
    dest="dump",
    default=os.devnull,
    help="File where to dump errors (default: redirect to devnull)",
)
parser.add_argument(
    "-s",
    "--context-size",
    dest="context_size",
    type=int,
    default=30,
    help="context size (default: %(default)s)",
)



# -*- coding: utf-8 -*-

"""
file: gui.py

Description: performs a sequence of operations in a pipeline.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import tkinter
from tkinter import ttk
import tkinter.messagebox

import time
import os
import argparse

import sem
import sem.importers
from sem.modules.tagger import tagger
from sem.gui_components import (
    SemTkResourceSelector,
    SemTkLangSelector,
    SemTkFileSelector,
    SemTkExportSelector,
    SEMTkTrainInterface,
)
import sem.logger
import sem.pipelines


class SemTkMainWindow(ttk.Frame):
    def __init__(self, root, resource_dir):
        """
        create the main window.
        """

        ttk.Frame.__init__(self, root)

        self.resource_dir = resource_dir

        self.current_files = None
        self.current_output = sem.SEM_DATA_DIR / "outputs"
        if not self.current_output.exists():
            os.makedirs(self.current_output)

        self.workflow_zone = ttk.Frame(root)
        self.workflow_zone.grid(row=0, column=0, rowspan=2, columnspan=1, sticky="ns")
        self.workflow_zone.pack(side=tkinter.LEFT, fill=tkinter.BOTH, expand=True)

        self.file_select_zone = ttk.Frame(root)
        self.file_select_zone.pack(side=tkinter.LEFT, fill=tkinter.BOTH, expand=True)

        self.launch_zone = ttk.Frame(root)
        self.launch_zone.pack(side=tkinter.LEFT, fill=tkinter.BOTH, expand=True)

        self.lang_selector = SemTkLangSelector(self.workflow_zone, self.resource_dir / "workflow")
        self.workflow = SemTkResourceSelector(self.workflow_zone, self.resource_dir / "workflow")
        self.lang_selector.register(self.workflow)
        self.lang_selector.pack()
        self.workflow.pack()

        self.file_selector = SemTkFileSelector(
            self.file_select_zone, self, button_opt={"fill": "both", "padx": 5, "pady": 5}
        )
        self.file_selector.pack()
        root.bind("<Control-o>", self.file_selector.filenames)

        self.export_format_selector = SemTkExportSelector(self.launch_zone)
        self.export_format_selector.pack()

        self.launch_button = ttk.Button(
            self.launch_zone, text="launch SEM", command=self.launch_tagger
        )
        self.launch_button.pack(expand=True)
        self.haw = tkinter.PhotoImage(file=str(self.resource_dir / "images" / "haw_24_24.gif"))
        self.launch_button.config(image=self.haw, compound=tkinter.LEFT)

        self.train_button = ttk.Button(
            self.launch_zone, text="train SEM", command=self.train_tagger
        )
        self.train_button.pack(expand=True)
        self.university = tkinter.PhotoImage(
            file=str(self.resource_dir / "images" / "fa_university_24_24.gif")
        )
        self.train_button.config(image=self.university, compound=tkinter.LEFT)

    def launch_tagger(self):
        current_files = self.file_selector.files()
        if not current_files:
            tkinter.messagebox.showwarning("launching SEM", "No files specified.")
            return

        workflow = self.workflow.resource()
        if not workflow:
            tkinter.messagebox.showwarning("launching SEM", "No workflow selected.")
            return
        workflowfile = self.resource_dir / "workflow" / self.lang_selector.lang() / workflow

        output_dir = self.current_output / time.strftime("%Y%m%d%H%M%S")
        if not output_dir.exists():
            os.makedirs(output_dir)

        try:
            export_format = self.export_format_selector.export_format()
            args = {
                "workflow": workflowfile,
                "infiles": [],
                "output_directory": output_dir,
                "force_format": export_format,
                "n_procs": 0,
            }
            for current_file in current_files:
                corpus = None
                try:
                    corpus = sem.importers.sem_document_from_xml(current_file)
                except Exception:
                    pass
                if corpus is not None:
                    args["infiles"].extend(corpus.documents)
                else:
                    args["infiles"].append(current_file)
            tagger(**args)
        except Exception as e:
            # handling ExpatError from etree
            tkinter.messagebox.showerror("launching SEM", "Error: {}".format(e))
            raise
        sem.logger.info("files are located in: {}".format(output_dir))
        tkinter.messagebox.showinfo(
            "launching SEM", "Everything went ok! files are located in: {}".format(output_dir)
        )
        return

    def train_tagger(self, event=None):
        current_files = self.file_selector.files()
        if not current_files:
            tkinter.messagebox.showwarning("launching SEM", "No files specified.")
            return

        SEMTkTrainInterface(self.file_selector, None, None)


def main(argv=None):
    gui(**vars(parser.parse_args(argv)))


def gui(resources=sem.SEM_RESOURCE_DIR, log_level="INFO"):
    root = tkinter.Tk()
    root.title("SEM")
    root.minsize(width=380, height=200)
    sem.logger.setLevel(log_level)

    SemTkMainWindow(root, resources).pack()

    root.mainloop()


parser = argparse.ArgumentParser(
    "Launch GUI for SEM for tagging documents and training new models."
)

parser.add_argument(
    "resources",
    nargs="?",
    default=sem.SEM_RESOURCE_DIR,
    help="The directory where resources are. If not provided,"
    " will use current installation's resources.",
)
parser.add_argument(
    "-l",
    "--log",
    dest="log_level",
    choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
    default="INFO",
    help="Increase log level (default: %(default)s)",
)



# -*- coding:utf-8-*-

"""
file: CRF.py

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re
import itertools

UNICODE_LOWERS = (
    "a-zµßàáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿāăąćĉċčďđēĕėęěĝğġģĥħĩīĭįıĳĵķĸĺļľŀłńņňŉŋōŏőœŕŗřśŝşšţťŧũūŭů"
    "űųŵŷźżžſƀƃƅƈƌƍƒƕƙƚƛƞơƣƥƨƪƫƭưƴƶƹƺƽƾƿǆǉǌǎǐǒǔǖǘǚǜǝǟǡǣǥǧǩǫǭǯǰǳǵǹǻǽǿȁȃȅȇȉȋȍȏȑȓȕȗșțȝȟȡȣȥȧȩȫȭȯȱȳȴȵ"
    "ȶȸȹȼȿɀɂɇɉɋɍɏɐɑɒɓɔɕɖɗɘəɚɛɜɝɞɟɠɡɢɣɤɥɦɧɨɩɪɫɬɭɮɯɰɱɲɳɴɵɶɷɸɹɺɻɼɽɾɿʀʁʂʃʄʅʆʇʈʉʊʋʌʍʎʏʐʑʒʓʕʖʗʘʙʚʛʜʝʞʟʠʡʢ"
    "ʣʤʥʦʧʨʩʪʫʬʭʮʯͻͼͽΐάέήίΰαβγδεζηθικλμνξοπρςστυφχψωϊϋόύώϐϑϕϖϗϙϛϝϟϡϣϥϧϩϫϭϯϰϱϲϳϵϸϻϼабвгдежзийклмноп"
    "рстуфхцчшщъыьэюяѐёђѓєѕіїјљњћќѝўџѡѣѥѧѩѫѭѯѱѳѵѷѹѻѽѿҁҋҍҏґғҕҗҙқҝҟҡңҥҧҩҫҭүұҳҵҷҹһҽҿӂӄӆӈӊӌӎӏӑӓӕ"
    "ӗәӛӝӟӡӣӥӧөӫӭӯӱӳӵӷӹӻӽӿԁԃԅԇԉԋԍԏԑԓԛԝաբգդեզէըթժիլխծկհձղճմյնշոչպջռսվտրցւփքօֆևᴀᴁᴂᴃᴄᴅᴆᴇᴈᴉᴊᴋᴌᴍᴎᴏᴐᴑᴒ"
    "ᴓᴔᴕᴖᴗᴘᴙᴚᴛᴜᴝᴞᴟᴠᴡᴢᴣᴤᴥᴦᴧᴨᴩᴪᴫᵫᵬᵭᵮᵯᵰᵱᵲᵳᵴᵵᵶᵷᵹᵺᵻᵼᵽᵾᵿᶀᶁᶂᶃᶄᶅᶆᶇᶈᶉᶊᶋᶌᶍᶎᶏᶐᶑᶒᶓᶔᶕᶖᶗᶘᶙᶚḁḃḅḇḉḋḍḏḑḓḕḗḙḛḝḟḡḣḥḧḩ"
    "ḫḭḯḱḳḵḷḹḻḽḿṁṃṅṇṉṋṍṏṑṓṕṗṙṛṝṟṡṣṥṧṩṫṭṯṱṳṵṷṹṻṽṿẁẃẅẇẉẋẍẏẑẓẕẗẘẙẚẛạảấầẩẫậắằẳẵặẹẻẽếềểễệỉịọỏốồổỗộớờởỡợ"
    "ụủứừửữựỳỵỷỹἀἁἂἃἄἅἆἇἐἑἒἓἔἕἠἡἢἣἤἥἦἧἰἱἲἳἴἵἶἷὀὁὂὃὄὅὐὑὒὓὔὕὖὗὠὡὢὣὤὥὦὧὰάὲέὴήὶίὸόὺύὼώᾀᾁᾂᾃᾄᾅᾆᾇᾐᾑᾒᾓᾔᾕᾖᾗ"
    "ᾠᾡᾢᾣᾤᾥᾦᾧᾰᾱᾲᾳᾴᾶᾷιῂῃῄῆῇῐῑῒΐῖῗῠῡῢΰῤῥῦῧῲῳῴῶῷℓⅎↄⱡⱥⱦⱨⱪⱬⱱⱳⱴⱶⱷ"
)
UNICODE_UPPERS = (
    "A-ZÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞĀĂĄĆĈĊČĎĐĒĔĖĘĚĜĞĠĢĤĦĨĪĬĮİĲĴĶĹĻĽĿŁŃŅŇŊŌŎŐŒŔŖŘŚŜŞŠŢŤŦŨŪŬŮŰŲŴŶŸ"
    "ŹŻŽƁƂƄƆƇƉƊƋƎƏƐƑƓƔƖƗƘƜƝƟƠƢƤƦƧƩƬƮƯƱƲƳƵƷƸƼǄǇǊǍǏǑǓǕǗǙǛǞǠǢǤǦǨǪǬǮǱǴǶǷǸǺǼǾȀȂȄȆȈȊȌȎȐȒȔȖȘȚȜȞȠȢȤȦȨȪȬȮȰȲ"
    "ȺȻȽȾɁɃɄɅɆɈɊɌɎΆΈΉΊΌΎΏΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩΪΫϏϒϓϔϘϚϜϞϠϢϤϦϨϪϬϮϴϷϹϺϽϾϿЀЁЂЃЄЅІЇЈЉЊЋЌЍЎЏАБВГДЕЖЗ"
    "ИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯѠѢѤѦѨѪѬѮѰѲѴѶѸѺѼѾҀҊҌҎҐҒҔҖҘҚҜҞҠҢҤҦҨҪҬҮҰҲҴҶҸҺҼҾӀӁӃӅӇӉӋӍӐӒӔӖӘӚӜӞӠӢӤӦӨӪӬӮӰ"
    "ӲӴӶӸӺӼӾԀԂԄԆԈԊԌԎԐԒԚԜԱԲԳԴԵԶԷԸԹԺԻԼԽԾԿՀՁՂՃՄՅՆՇՈՉՊՋՌՍՎՏՐՑՒՓՔՕՖᎠᎡᎢᎣᎤᎥᎦᎧᎨᎩᎪᎫᎬᎭᎮᎯᎰᎱᎲᎳᎴᎵᎶᎷᎸᎹᎺᎻᎼᎽᎾᎿᏀᏁᏂᏃ"
    "ᏄᏅᏆᏇᏈᏉᏊᏋᏌᏍᏎᏏᏐᏑᏒᏓᏔᏕᏖᏗᏘᏙᏚᏛᏜᏝᏞᏟᏠᏡᏢᏣᏤᏥᏦᏧᏨᏩᏪᏫᏬᏭᏮᏯᏰᏱᏲᏳᏴᏵḀḂḄḆḈḊḌḎḐḒḔḖḘḚḜḞḠḢḤḦḨḪḬḮḰḲḴḶḸḺḼḾṀṂṄṆṈṊṌṎṐṒṔ"
    "ṖṘṚṜṞṠṢṤṦṨṪṬṮṰṲṴṶṸṺṼṾẀẂẄẆẈẊẌẎẐẒẔẞẠẢẤẦẨẪẬẮẰẲẴẶẸẺẼẾỀỂỄỆỈỊỌỎỐỒỔỖỘỚỜỞỠỢỤỦỨỪỬỮỰỲỴỶỸἈἉἊἋἌἍἎἏἘἙἚἛἜἝἨ"
    "ἩἪἫἬἭἮἯἸἹἺἻἼἽἾἿὈὉὊὋὌὍὙὛὝὟὨὩὪὫὬὭὮὯᾸᾹᾺΆῈΈῊΉῘῙῚΊῨῩῪΎῬῸΌῺΏⱠⱢⱣⱤⱧⱩⱫⱭⱲⱵＡＢＣＤＥＦＧＨＩＪＫＬＭＮＯＰＱＲＳＴＵＶＷＸＹＺ𐐀𐐁𐐂"
    "𐐃𐐄𐐅𐐆𐐇𐐈𐐉𐐊𐐋𐐌𐐍𐐎𐐏𐐐𐐑𐐒𐐓𐐔𐐕𐐖𐐗𐐘𐐙𐐚𐐛𐐜𐐝𐐞𐐟𐐠𐐡𐐢𐐣𐐤𐐥𐐦𐐧"
)
UNICODE_DIGITS = "0-9"
UNICODE_PUNCTS = (
    "-֊־᠆‒–—―⸗〜〰゠︱﹣－_︳︴﹍﹎﹏＿\\)\\]\\}᚜〉》」』】〕〗〙〛〞〟﴾︘︶︸︺︼︾﹀﹂﹄﹚﹜﹞）］｝｠｣\\(\\[\\{᚛‚„〈《「『【〔〖〘〚〝﴿︗︵︷︹︻︽︿﹁﹃"
    "﹙﹛﹝（［｛｟｢!\"#%&'*,./:;?@\\\\¡§¶·¿;·՚՛՜՝՞՟։׀׃׆׳״،؍؛؞؟٪٫٬٭۔܀܁܂܃܄܅܆܇܈܉܊܋܌܍߷߸߹।॥॰෴๏๚๛༄༅༆༇༈༉༊་༌།༎༏༐"
    "༑༒༔྅࿐࿑࿒࿓࿔჻፠፡።፣፤፥፦፧፨᙭᙮᛫᛬᛭។៕៖៘៙៚᠀᠁᠂᠃᠄᠅᠇᠈᠉᠊‗†‡•…‰′″‴※‼‾⁞、。〃〽・꘍꘎꘏꡴꡵꡶꡷︐︑︒︓︔︕︖︙︰﹅﹆﹉﹊﹋﹌﹐﹑﹒﹔﹕﹖﹗﹟﹠﹡﹨﹪﹫"
    "！＂＃％＆＇＊，．／：；？＠＼｡､･«»•"
)
UNICODE_ALPHAS = UNICODE_LOWERS + UNICODE_UPPERS
UNICODE_ALPHANUMS = UNICODE_ALPHAS + UNICODE_DIGITS


class Pattern:
    def __init__(self, case_insensitive=False, *args):
        self._case_insensitive = case_insensitive
        pass

    def instanciate(self, matrix, index):
        raise RuntimeError("undefined")


class ConstantPattern(Pattern):
    def __init__(self, value, case_insensitive=False, *args):
        self._value = value

    def __str__(self):
        return self._value

    def instanciate(self, matrix, index):
        return self._value

    @property
    def value(self):
        return self._value


class IdentityPattern(Pattern):
    __pattern = re.compile("%x\\[(-?[0-9]+),([0-9]+)\\]")

    def __init__(self, x, y, case_insensitive=False, column=None, *args):
        self._x = x
        self._y = y
        self._column = column

    def __str__(self):
        return "%x[{p.x},{p.y}]".format(p=self)

    def instanciate(self, matrix, index):
        x = index + self.x
        if x < 0:
            if x > -5:
                return "_x{0:+d}".format(x)
            else:
                return "_x-#"
        if x >= len(matrix):
            diff = x - len(matrix) + 1
            if diff < 5:
                return "_x{0:+d}".format(diff)
            else:
                return "_x+#"

        return str(matrix[x][self.y])

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def column(self):
        return self._column

    @classmethod
    def from_string(cls, string, case_insensitive=False, column=None):
        match = cls.__pattern.search(string)
        groups = match.groups()
        return IdentityPattern(
            int(groups[0]), int(groups[1]), case_insensitive=case_insensitive, column=column
        )


class RegexPattern(IdentityPattern):
    __sub = {
        "\\l": "[{0}]".format(UNICODE_LOWERS),
        "\\L": "[^{0}]".format(UNICODE_LOWERS),
        "\\u": "[{0}]".format(UNICODE_UPPERS),
        "\\U": "[^{0}]".format(UNICODE_UPPERS),
        "\\d": "[{0}]".format(UNICODE_DIGITS),
        "\\D": "[^{0}]".format(UNICODE_DIGITS),
        "\\p": "[{0}]".format(UNICODE_PUNCTS),
        "\\P": "[^{0}]".format(UNICODE_PUNCTS),
        "\\a": "[{0}]".format(UNICODE_ALPHAS),
        "\\A": "[^{0}]".format(UNICODE_ALPHAS),
        "\\w": "[{0}]".format(UNICODE_ALPHANUMS),
        "\\W": "[^{0}]".format(UNICODE_ALPHANUMS),
    }

    def __init__(self, x, y, pattern, case_insensitive=False, *args):
        super(RegexPattern, self).__init__(x, y, *args)
        self._pattern = pattern
        for key, value in RegexPattern.__sub.items():
            self._pattern = self._pattern.replace(key, value)
        self._pattern = re.compile(self._pattern, re.U + re.M)

    @classmethod
    def sub(cls):
        return RegexPattern.__sub

    """def instanciate(self, matrix, index):
        cell = super(RegexPattern, self).instanciate(matrix, index)

        return self._pattern.search(cell)"""


class TestPattern(RegexPattern):
    __pattern = re.compile(r'%t\[\s*(-?[0-9]+),([0-9]+),"(.+)"\]', re.I)

    def __str__(self):
        p = self._pattern.pattern[:]
        for x, y in RegexPattern.sub().items():
            p = p.replace(y, x)
        return '%t[{p.x},{p.y},"{pat}"]'.format(p=self, pat=p)

    def instanciate(self, matrix, index, case_insensitive=False):
        cell = super(TestPattern, self).instanciate(matrix, index)

        return (self._pattern.search(cell) is not None).lower()

    @classmethod
    def from_string(cls, string, case_insensitive=False, column=None):
        match = cls.__pattern.search(string)
        groups = match.groups()
        return TestPattern(
            int(groups[0]), int(groups[1]), groups[2], case_insensitive=case_insensitive
        )


class MatchPattern(RegexPattern):
    __pattern = re.compile(r'%m\[\s*(-?[0-9]+),([0-9]+),"([^"]+?)"\]', re.I)

    def __str__(self):
        p = self._pattern.pattern[:]
        for x, y in RegexPattern.sub().items():
            p = p.replace(y, x)
        return '%m[{p.x},{p.y},"{pat}"]'.format(p=self, pat=p)

    def instanciate(self, matrix, index, case_insensitive=False):
        cell = super(MatchPattern, self).instanciate(matrix, index)

        cell = self._pattern.search(cell)
        if cell is None:
            return ""

        return cell.group()

    @classmethod
    def from_string(cls, string, case_insensitive=False, column=None):
        match = cls.__pattern.search(string)
        groups = match.groups()
        return MatchPattern(
            int(groups[0]), int(groups[1]), groups[2], case_insensitive=case_insensitive
        )


class ListPattern(Pattern):
    __pattern = re.compile(r'%[xtm]\[\s*-?[0-9]+,-?[0-9]+(,".+?")?\]', re.I)

    def __init__(self, patterns):
        self._patterns = patterns[:]
        self._instanciators = [pattern.instanciate for pattern in self._patterns]

    def __str__(self):
        return "".join([str(p) for p in self._patterns])

    @property
    def patterns(self):
        return self._patterns

    def instanciate(self, matrix, index):
        return "".join([instanciator(matrix, index) for instanciator in self._instanciators])

    @classmethod
    def from_string(cls, string, case_insensitive=False, column=None):
        patterns = []
        prev = 0
        finding = None
        for finding in cls.__pattern.finditer(string):
            patterns.append(pattern_factory(string[prev: finding.start()]))
            patterns.append(pattern_factory(string[finding.start(): finding.end()]))
            prev = finding.end()
        if finding is None:
            patterns.append(ConstantPattern(string))
        elif string[prev:]:
            patterns.append(ConstantPattern(string[prev:]))
        return ListPattern(patterns)


def pattern_factory(string):
    if string.startswith("%x"):
        return IdentityPattern.from_string(
            string, case_insensitive=string[1].isupper(), column=None
        )
    elif string.startswith("%t"):
        return TestPattern.from_string(string, case_insensitive=string[1].isupper(), column=None)
    elif string.startswith("%m"):
        return MatchPattern.from_string(string, case_insensitive=string[1].isupper(), column=None)
    return ConstantPattern(string)


class Quark:
    def __init__(self):
        self._encoder = {}
        self._decoder = []

    def __len__(self):
        return len(self._decoder)

    def __iter__(self):
        return iter(self._decoder[:])

    def __contains__(self, element):
        return element in self._encoder

    def keys(self):
        return self._decoder[:]

    def add(self, element, strict=False):
        if element not in self._encoder:
            self._encoder[element] = len(self)
            self._decoder.append(element)
        elif strict:
            raise KeyError("'{0}' already in coder".format(element))

    def insert(self, index, element):
        if element not in self._encoder:
            self._decoder.insert(index, element)
            self._encoder[element] = index
            for nth in range(index + 1, len(self._encoder)):
                self._encoder[self._decoder[nth]] = nth

    def encode(self, element):
        return self._encoder.get(element, -1)

    def decode(self, integer):
        try:
            return self._decoder[integer]
        except Exception:
            return None


class Model:
    def __init__(self, constraints={}):
        self._tagset = Quark()
        self._templates = []
        self._observations = Quark()
        self._uoff = []
        self._boff = []
        self._weights = []  # list
        self._max_col = 0

    def __call__(self, x):
        return self.tag_viterbi(x)

    @classmethod
    def from_wapiti_model(cls, filename, encoding="utf-8", verbose=True):
        model = Model()
        n_weights = -1
        n_patterns = -1
        n_labels = -1
        n_observations = -1
        current_feature = 0
        line_index = 0
        mode = ("rb" if encoding is None else "r")
        with open(filename, mode, encoding=encoding) as input_stream:
            lines = [line.strip() for line in input_stream]

        n_weights = int(lines[line_index].split("#")[-1])
        line_index += 1

        n_patterns, max_col, other = lines[line_index].split("#")[-1].split("/")
        n_patterns = int(n_patterns)
        max_col = int(max_col)
        model._max_col = max_col
        line_index += 1

        for line in lines[line_index: line_index + n_patterns]:
            line = line.split(":", 1)[1][:-1]
            model._templates.append(ListPattern.from_string(line))
        assert len(model.templates) == n_patterns
        line_index += n_patterns

        # labels
        n_labels = int(lines[line_index].strip().split("#")[-1])
        line_index += 1
        for line in lines[line_index: line_index + n_labels]:
            line = line.strip()
            model.tagset.add(line[line.index(":") + 1: -1])
        line_index += n_labels
        assert len(model.tagset) == n_labels

        # observations
        n_observations = int(lines[line_index].split("#")[-1])
        line_index += 1
        model._uoff = [-1] * n_observations
        model._boff = [-1] * n_observations
        for line in lines[line_index: line_index + n_observations]:
            obs = line[line.index(":") + 1: -1]
            n_feats = n_labels if obs[0] in "u*" else 0
            n_feats += n_labels ** 2 if obs[0] in "b*" else 0

            model.observations.add(obs)
            if obs[0] in "u*":
                model.uoff[len(model.observations) - 1] = current_feature
            if obs[0] in "b*":
                model.boff[len(model.observations) - 1] = current_feature
                +(n_labels if obs[0] == "*" else 0)
            current_feature += n_feats
        line_index += n_observations
        model._weights = [0.0] * current_feature

        # features
        for line in lines[-n_weights:]:
            index, weight = line.split("=")
            model.weights[int(index)] = float.fromhex(weight)

        return model

    @property
    def tagset(self):
        return self._tagset

    @property
    def templates(self):
        return self._templates

    @property
    def observations(self):
        return self._observations

    @property
    def uoff(self):
        """Unigram OFFset"""
        return self._uoff

    @property
    def boff(self):
        """Bigram OFFset"""
        return self._boff

    @property
    def weights(self):
        return self._weights

    def tag_viterbi(self, sentence):
        Y = len(self.tagset)
        T = len(sentence)
        range_Y = range(Y)
        range_T = range(T)
        psi = [[[0.0] * Y for _y1 in range_Y] for _t in range_T]
        back = [[0] * Y for _t in range_T]
        cur = [0.0] * Y
        old = [0.0] * Y
        psc = [0.0] * T
        sc = -2 ** 30
        tag = ["" for _t in range_T]
        # avoiding dots
        weights_ = self._weights
        obs_encode = self._observations.encode
        templates_ = self._templates
        uoff_ = self._uoff
        boff_ = self._boff

        unigrams = []
        bigrams = []
        for t in range_T:
            unigrams.append([])
            bigrams.append([])
            u_append = unigrams[-1].append
            b_append = bigrams[-1].append
            for template in templates_:
                obs = template.instanciate(sentence, t)
                o = obs_encode(obs)
                if o != -1:
                    if obs[0] == "u" and uoff_[o] != -1:
                        u_append(weights_[uoff_[o]: uoff_[o] + Y])
                    if obs[0] == "b" and boff_[o] != -1:
                        b_append(weights_[boff_[o]: boff_[o] + Y * Y])

        # compute scores in psi
        for t in range_T:
            unigrams_T = unigrams[t]
            for y in range_Y:
                sum_ = 0.0
                for w in unigrams_T:
                    sum_ += w[y]
                for yp in range_Y:
                    psi[t][yp][y] = sum_
        for t in range(1, T):
            bigrams_T = bigrams[t]
            d = 0
            for yp, y in itertools.product(range_Y, range_Y):
                for w in bigrams_T:
                    psi[t][yp][y] += w[d]
                d += 1

        for y in range_Y:
            cur[y] = psi[0][0][y]
        for t in range(1, T):
            for y in range_Y:
                old[y] = cur[y]
            for y in range_Y:
                bst = -2 ** 30
                idx = 0
                for yp in range_Y:
                    val = old[yp] + psi[t][yp][y]
                    if val > bst:
                        bst = val
                        idx = yp
                back[t][y] = idx
                cur[y] = bst

        bst = 0
        for y in range(1, Y):
            if cur[y] > cur[bst]:
                bst = y
        sc = cur[bst]
        for t in reversed(range_T):
            yp = back[t][bst] if t != 0 else 0
            y = bst
            tag[t] = self._tagset.decode(y)
            psc[t] = psi[t][yp][y]
            bst = yp

        return tag, psc, sc

    def write(self, filename, encoding="utf-8"):
        with open(filename, "w", encoding=encoding, newline="") as output_stream:
            output_stream.write("#mdl#2#{0}\n".format(len([w for w in self.weights if w != 0.0])))
            output_stream.write("#rdr#{0}/{1}/0\n".format(len(self._templates), self._max_col))
            for pattern in self._templates:
                uni_pattern = str(pattern)
                output_stream.write("{0}:{1},\n".format(len(uni_pattern), uni_pattern))
            output_stream.write("#qrk#{0}\n".format(len(self._tagset)))
            for tag in self.tagset:
                output_stream.write("{0}:{1}\n".format(len(tag), tag))
            # observations
            output_stream.write("#qrk#{0}\n".format(len(self._observations)))
            for obs in self._observations:
                output_stream.write("{0}:{1}\n".format(len(obs), obs))
            for index, w in enumerate(self.weights):
                if w != 0.0:
                    output_stream.write("{0}={1}\n".format(index, float.hex(w)))

    def dump(self, filename):
        ntags = len(self.tagset)
        with open(filename, "w", encoding="utf-8", newline="") as output_stream:
            for i in range(len(self.observations)):
                o = self.observations.decode(i)
                written = False
                if o[0] == "u":
                    off = self.uoff[i]
                    for y in range(ntags):
                        w = self.weights[off + y]
                        if w != 0:
                            output_stream.write(
                                "{0}\t{1}\t{2}\t{3:.5f}\n".format(o, "#", self.tagset.decode(y), w)
                            )
                            written = True
                else:
                    off = self.boff[i]
                    d = 0
                    for yp in range(ntags):
                        for y in range(ntags):
                            w = self.weights[off + d]
                            if w != 0:
                                output_stream.write(
                                    "{0}\t{1}\t{2}\t{3:.5f}\n".format(
                                        o, self.tagset.decode(yp), self.tagset.decode(y), w
                                    )
                                )
                                written = True
                            d += 1
                if written:
                    output_stream.write("\n")
